@@ -75,13 +75,6 @@ typedef struct {
     int         full_sweep;
 } jl_gc_num_t;
 
-typedef struct {
-    void **pc; // Current stack address for the pc (up growing)
-    jl_gc_mark_data_t *data; // Current stack address for the data (up growing)
-    void **pc_start; // Cached value of `gc_cache->pc_stack`
-    void **pc_end; // Cached value of `gc_cache->pc_stack_end`
-} gc_mark_sp_t;
-
 enum {
     GC_MARK_L_marked_obj,
     GC_MARK_L_scan_only,
@@ -91,19 +84,21 @@ enum {
     GC_MARK_L_obj16,
     GC_MARK_L_obj32,
     GC_MARK_L_stack,
+    GC_MARK_L_excstack,
     GC_MARK_L_module_binding,
     _GC_MARK_L_MAX
 };
 
-/**
- * The `nptr` member of marking data records the number of pointers slots referenced by
- * an object to be used in the full collection heuristics as well as whether the object
- * references young objects.
- * `nptr >> 2` is the number of pointers fields referenced by the object.
- * The lowest bit of `nptr` is set if the object references young object.
- * The 2nd lowest bit of `nptr` is the GC old bits of the object after marking.
- * A `0x3` in the low bits means that the object needs to be in the remset.
- */
+// The following structs (`gc_mark_*_t`) contain iterator state used for the
+// scanning of various object types.
+//
+// The `nptr` member records the number of pointers slots referenced by
+// an object to be used in the full collection heuristics as well as whether the object
+// references young objects.
+// `nptr >> 2` is the number of pointers fields referenced by the object.
+// The lowest bit of `nptr` is set if the object references young object.
+// The 2nd lowest bit of `nptr` is the GC old bits of the object after marking.
+// A `0x3` in the low bits means that the object needs to be in the remset.
 
 // An generic object that's marked and needs to be scanned
 // The metadata might need update too (depend on the PC)
@@ -156,6 +151,13 @@ typedef struct {
     uintptr_t ub;
 } gc_mark_stackframe_t;
 
+// Exception stack data
+typedef struct {
+    jl_excstack_t *s;  // Stack of exceptions
+    size_t itr;        // Iterator into exception stack
+    size_t i;          // Iterator into backtrace data for exception
+} gc_mark_excstack_t;
+
 // Module bindings. This is also the beginning of module scanning.
 // The loop will start marking other references in a module after the bindings are marked
 typedef struct {
@@ -183,6 +185,7 @@ union _jl_gc_mark_data {
     gc_mark_obj16_t obj16;
     gc_mark_obj32_t obj32;
     gc_mark_stackframe_t stackframe;
+    gc_mark_excstack_t excstackframe;
     gc_mark_binding_t binding;
     gc_mark_finlist_t finlist;
 };
@@ -190,7 +193,7 @@ union _jl_gc_mark_data {
 // Pop a data struct from the mark data stack (i.e. decrease the stack pointer)
 // This should be used after dispatch and therefore the pc stack pointer is already popped from
 // the stack.
-STATIC_INLINE void *gc_pop_markdata_(gc_mark_sp_t *sp, size_t size)
+STATIC_INLINE void *gc_pop_markdata_(jl_gc_mark_sp_t *sp, size_t size)
 {
     jl_gc_mark_data_t *data = (jl_gc_mark_data_t *)(((char*)sp->data) - size);
     sp->data = data;
@@ -201,7 +204,7 @@ STATIC_INLINE void *gc_pop_markdata_(gc_mark_sp_t *sp, size_t size)
 // Re-push a frame to the mark stack (both data and pc)
 // The data and pc are expected to be on the stack (or updated in place) already.
 // Mainly useful to pause the current scanning in order to scan an new object.
-STATIC_INLINE void *gc_repush_markdata_(gc_mark_sp_t *sp, size_t size)
+STATIC_INLINE void *gc_repush_markdata_(jl_gc_mark_sp_t *sp, size_t size) JL_NOTSAFEPOINT
 {
     jl_gc_mark_data_t *data = sp->data;
     sp->pc++;
@@ -348,6 +351,9 @@ typedef struct {
     int ub;
 } pagetable_t;
 
+#ifdef __clang_analyzer__
+unsigned ffs_u32(uint32_t bitvec) JL_NOTSAFEPOINT;
+#else
 STATIC_INLINE unsigned ffs_u32(uint32_t bitvec)
 {
 #if defined(_COMPILER_MINGW_)
@@ -360,6 +366,7 @@ STATIC_INLINE unsigned ffs_u32(uint32_t bitvec)
     return ffs(bitvec) - 1;
 #endif
 }
+#endif
 
 extern jl_gc_num_t gc_num;
 extern pagetable_t memory_map;
@@ -368,55 +375,55 @@ extern arraylist_t finalizer_list_marked;
 extern arraylist_t to_finalize;
 extern int64_t lazy_freed_pages;
 
-STATIC_INLINE bigval_t *bigval_header(jl_taggedvalue_t *o)
+STATIC_INLINE bigval_t *bigval_header(jl_taggedvalue_t *o) JL_NOTSAFEPOINT
 {
     return container_of(o, bigval_t, header);
 }
 
 // round an address inside a gcpage's data to its beginning
-STATIC_INLINE char *gc_page_data(void *x)
+STATIC_INLINE char *gc_page_data(void *x) JL_NOTSAFEPOINT
 {
     return (char*)(((uintptr_t)x >> GC_PAGE_LG2) << GC_PAGE_LG2);
 }
 
-STATIC_INLINE jl_taggedvalue_t *page_pfl_beg(jl_gc_pagemeta_t *p)
+STATIC_INLINE jl_taggedvalue_t *page_pfl_beg(jl_gc_pagemeta_t *p) JL_NOTSAFEPOINT
 {
     return (jl_taggedvalue_t*)(p->data + p->fl_begin_offset);
 }
 
-STATIC_INLINE jl_taggedvalue_t *page_pfl_end(jl_gc_pagemeta_t *p)
+STATIC_INLINE jl_taggedvalue_t *page_pfl_end(jl_gc_pagemeta_t *p) JL_NOTSAFEPOINT
 {
     return (jl_taggedvalue_t*)(p->data + p->fl_end_offset);
 }
 
-STATIC_INLINE int gc_marked(uintptr_t bits)
+STATIC_INLINE int gc_marked(uintptr_t bits) JL_NOTSAFEPOINT
 {
     return (bits & GC_MARKED) != 0;
 }
 
-STATIC_INLINE int gc_old(uintptr_t bits)
+STATIC_INLINE int gc_old(uintptr_t bits) JL_NOTSAFEPOINT
 {
     return (bits & GC_OLD) != 0;
 }
 
-STATIC_INLINE uintptr_t gc_set_bits(uintptr_t tag, int bits)
+STATIC_INLINE uintptr_t gc_set_bits(uintptr_t tag, int bits) JL_NOTSAFEPOINT
 {
     return (tag & ~(uintptr_t)3) | bits;
 }
 
-STATIC_INLINE uintptr_t gc_ptr_tag(void *v, uintptr_t mask)
+STATIC_INLINE uintptr_t gc_ptr_tag(void *v, uintptr_t mask) JL_NOTSAFEPOINT
 {
     return ((uintptr_t)v) & mask;
 }
 
-STATIC_INLINE void *gc_ptr_clear_tag(void *v, uintptr_t mask)
+STATIC_INLINE void *gc_ptr_clear_tag(void *v, uintptr_t mask) JL_NOTSAFEPOINT
 {
     return (void*)(((uintptr_t)v) & ~mask);
 }
 
 NOINLINE uintptr_t gc_get_stack_ptr(void);
 
-STATIC_INLINE jl_gc_pagemeta_t *page_metadata(void *_data)
+STATIC_INLINE jl_gc_pagemeta_t *page_metadata(void *_data) JL_NOTSAFEPOINT
 {
     uintptr_t data = ((uintptr_t)_data);
     unsigned i;
@@ -441,7 +448,7 @@ struct jl_gc_metadata_ext {
     unsigned pagetable0_i32, pagetable0_i;
 };
 
-STATIC_INLINE struct jl_gc_metadata_ext page_metadata_ext(void *_data)
+STATIC_INLINE struct jl_gc_metadata_ext page_metadata_ext(void *_data) JL_NOTSAFEPOINT
 {
     uintptr_t data = (uintptr_t)_data;
     struct jl_gc_metadata_ext info;
@@ -462,7 +469,7 @@ STATIC_INLINE struct jl_gc_metadata_ext page_metadata_ext(void *_data)
     return info;
 }
 
-STATIC_INLINE void gc_big_object_unlink(const bigval_t *hdr)
+STATIC_INLINE void gc_big_object_unlink(const bigval_t *hdr) JL_NOTSAFEPOINT
 {
     *hdr->prev = hdr->next;
     if (hdr->next) {
@@ -470,7 +477,7 @@ STATIC_INLINE void gc_big_object_unlink(const bigval_t *hdr)
     }
 }
 
-STATIC_INLINE void gc_big_object_link(bigval_t *hdr, bigval_t **list)
+STATIC_INLINE void gc_big_object_link(bigval_t *hdr, bigval_t **list) JL_NOTSAFEPOINT
 {
     hdr->next = *list;
     hdr->prev = list;
@@ -479,7 +486,7 @@ STATIC_INLINE void gc_big_object_link(bigval_t *hdr, bigval_t **list)
     *list = hdr;
 }
 
-STATIC_INLINE void gc_mark_sp_init(jl_gc_mark_cache_t *gc_cache, gc_mark_sp_t *sp)
+STATIC_INLINE void gc_mark_sp_init(jl_gc_mark_cache_t *gc_cache, jl_gc_mark_sp_t *sp)
 {
     sp->pc = gc_cache->pc_stack;
     sp->data = gc_cache->data_stack;
@@ -487,10 +494,11 @@ STATIC_INLINE void gc_mark_sp_init(jl_gc_mark_cache_t *gc_cache, gc_mark_sp_t *s
     sp->pc_end = gc_cache->pc_stack_end;
 }
 
-void gc_mark_queue_all_roots(jl_ptls_t ptls, gc_mark_sp_t *sp);
-void gc_mark_queue_finlist(jl_gc_mark_cache_t *gc_cache, gc_mark_sp_t *sp,
+void gc_mark_queue_all_roots(jl_ptls_t ptls, jl_gc_mark_sp_t *sp);
+void gc_mark_queue_finlist(jl_gc_mark_cache_t *gc_cache, jl_gc_mark_sp_t *sp,
                            arraylist_t *list, size_t start);
-void gc_mark_loop(jl_ptls_t ptls, gc_mark_sp_t sp);
+void gc_mark_loop(jl_ptls_t ptls, jl_gc_mark_sp_t sp);
+void sweep_stack_pools(void);
 void gc_debug_init(void);
 
 extern void *gc_mark_label_addrs[_GC_MARK_L_MAX];
@@ -498,8 +506,8 @@ extern void *gc_mark_label_addrs[_GC_MARK_L_MAX];
 // GC pages
 
 void jl_gc_init_page(void);
-NOINLINE jl_gc_pagemeta_t *jl_gc_alloc_page(void);
-void jl_gc_free_page(void *p);
+NOINLINE jl_gc_pagemeta_t *jl_gc_alloc_page(void) JL_NOTSAFEPOINT;
+void jl_gc_free_page(void *p) JL_NOTSAFEPOINT;
 
 // GC debug
 
@@ -615,7 +623,7 @@ extern int gc_verifying;
 #endif
 int gc_slot_to_fieldidx(void *_obj, void *slot);
 int gc_slot_to_arrayidx(void *_obj, void *begin);
-NOINLINE void gc_mark_loop_unwind(jl_ptls_t ptls, gc_mark_sp_t sp, int pc_offset);
+NOINLINE void gc_mark_loop_unwind(jl_ptls_t ptls, jl_gc_mark_sp_t sp, int pc_offset);
 
 #ifdef GC_DEBUG_ENV
 JL_DLLEXPORT extern jl_gc_debug_env_t jl_gc_debug_env;
@@ -623,7 +631,7 @@ JL_DLLEXPORT extern jl_gc_debug_env_t jl_gc_debug_env;
 int gc_debug_check_other(void);
 int gc_debug_check_pool(void);
 void gc_debug_print(void);
-void gc_scrub_record_task(jl_task_t *ta);
+void gc_scrub_record_task(jl_task_t *ta) JL_NOTSAFEPOINT;
 void gc_scrub(void);
 #else
 #define gc_sweep_always_full 0
@@ -638,7 +646,7 @@ static inline int gc_debug_check_pool(void)
 static inline void gc_debug_print(void)
 {
 }
-static inline void gc_scrub_record_task(jl_task_t *ta)
+static inline void gc_scrub_record_task(jl_task_t *ta) JL_NOTSAFEPOINT
 {
     (void)ta;
 }
@@ -648,11 +656,11 @@ static inline void gc_scrub(void)
 #endif
 
 #ifdef OBJPROFILE
-void objprofile_count(void *ty, int old, int sz);
+void objprofile_count(void *ty, int old, int sz) JL_NOTSAFEPOINT;
 void objprofile_printall(void);
 void objprofile_reset(void);
 #else
-static inline void objprofile_count(void *ty, int old, int sz)
+static inline void objprofile_count(void *ty, int old, int sz) JL_NOTSAFEPOINT
 {
 }
 
