@@ -17,6 +17,10 @@
 #include <getopt.h>
 #endif
 
+#if defined(_OS_FREEBSD_)
+#include <pthread_np.h>
+#endif
+
 #include "julia.h"
 #include "julia_internal.h"
 #define DEFINE_BUILTIN_GLOBALS
@@ -46,6 +50,9 @@ extern BOOL (WINAPI *hSymRefreshModuleList)(HANDLE);
 #include <unistd.h>
 #endif
 
+// list of modules being deserialized with __init__ methods
+jl_array_t *jl_module_init_order;
+
 #ifdef JL_ASAN_ENABLED
 JL_DLLEXPORT const char* __asan_default_options() {
     return "allow_user_segv_handler=1:detect_leaks=0";
@@ -62,20 +69,19 @@ void jl_init_stack_limits(int ismaster, void **stack_lo, void **stack_hi)
 #ifdef _OS_WINDOWS_
     (void)ismaster;
     // https://en.wikipedia.org/wiki/Win32_Thread_Information_Block
-#ifdef _P64
+#  ifdef _P64
     *stack_hi = (void**)__readgsqword(0x08); // Stack Base / Bottom of stack (high address)
     *stack_lo = (void**)__readgsqword(0x10); // Stack Limit / Ceiling of stack (low address)
-#else
+#  else // !_P64
     *stack_hi = (void**)__readfsdword(0x04); // Stack Base / Bottom of stack (high address)
     *stack_lo = (void**)__readfsdword(0x08); // Stack Limit / Ceiling of stack (low address)
-#endif
-#else
-#  ifdef JULIA_ENABLE_THREADING
+#  endif // _P64
+#else // !_OS_WINDOWS_
     // Only use pthread_*_np functions to get stack address for non-master
     // threads since it seems to return bogus values for master thread on Linux
     // and possibly OSX.
     if (!ismaster) {
-#    if defined(_OS_LINUX_)
+#  if defined(_OS_LINUX_)
         pthread_attr_t attr;
         pthread_getattr_np(pthread_self(), &attr);
         void *stackaddr;
@@ -83,18 +89,18 @@ void jl_init_stack_limits(int ismaster, void **stack_lo, void **stack_hi)
         pthread_attr_getstack(&attr, &stackaddr, &stacksize);
         pthread_attr_destroy(&attr);
         *stack_lo = (void*)stackaddr;
-        *stack_hi = (void*)((char*)stackaddr + stacksize);
+        *stack_hi = (void*)&stacksize;
         return;
-#    elif defined(_OS_DARWIN_)
+#  elif defined(_OS_DARWIN_)
         extern void *pthread_get_stackaddr_np(pthread_t thread);
         extern size_t pthread_get_stacksize_np(pthread_t thread);
         pthread_t thread = pthread_self();
         void *stackaddr = pthread_get_stackaddr_np(thread);
         size_t stacksize = pthread_get_stacksize_np(thread);
-        *stack_lo = (char*)stackaddr;
-        *stack_hi = (void*)((char*)stackaddr + stacksize);
+        *stack_lo = (void*)stackaddr;
+        *stack_hi = (void*)&stacksize;
         return;
-#    elif defined(_OS_FREEBSD_)
+#  elif defined(_OS_FREEBSD_)
         pthread_attr_t attr;
         pthread_attr_init(&attr);
         pthread_attr_get_np(pthread_self(), &attr);
@@ -102,21 +108,18 @@ void jl_init_stack_limits(int ismaster, void **stack_lo, void **stack_hi)
         size_t stacksize;
         pthread_attr_getstack(&attr, &stackaddr, &stacksize);
         pthread_attr_destroy(&attr);
-        *stack_lo = (char*)stackaddr;
-        *stack_hi = (void*)((char*)stackaddr + stacksize);
+        *stack_lo = (void*)stackaddr;
+        *stack_hi = (void*)&stacksize;
         return;
-#    else
-#      warning "Getting precise stack size for thread is not supported."
-#    endif
-    }
 #  else
-    (void)ismaster;
+#      warning "Getting precise stack size for thread is not supported."
 #  endif
+    }
     struct rlimit rl;
     getrlimit(RLIMIT_STACK, &rl);
-    size_t stack_size = rl.rlim_cur;
-    *stack_hi = (void*)&stack_size;
-    *stack_lo = (void*)((char*)*stack_hi - stack_size);
+    size_t stacksize = rl.rlim_cur;
+    *stack_hi = (void*)&stacksize;
+    *stack_lo = (void*)((char*)*stack_hi - stacksize);
 #endif
 }
 
@@ -153,8 +156,10 @@ static void jl_uv_exitcleanup_add(uv_handle_t *handle, struct uv_shutdown_queue
     struct uv_shutdown_queue_item *item = (struct uv_shutdown_queue_item*)malloc(sizeof(struct uv_shutdown_queue_item));
     item->h = handle;
     item->next = NULL;
-    if (queue->last) queue->last->next = item;
-    if (!queue->first) queue->first = item;
+    if (queue->last)
+        queue->last->next = item;
+    if (!queue->first)
+        queue->first = item;
     queue->last = item;
 }
 
@@ -216,6 +221,9 @@ static void jl_close_item_atexit(uv_handle_t *handle)
 
 JL_DLLEXPORT void jl_atexit_hook(int exitcode)
 {
+    if (jl_all_tls_states == NULL)
+        return;
+
     jl_ptls_t ptls = jl_get_ptls_states();
 
     if (exitcode == 0)
@@ -255,6 +263,7 @@ JL_DLLEXPORT void jl_atexit_hook(int exitcode)
     }
 
     struct uv_shutdown_queue queue = {NULL, NULL};
+    JL_UV_LOCK();
     uv_walk(loop, jl_uv_exitcleanup_walk, &queue);
     struct uv_shutdown_queue_item *item = queue.first;
     if (ptls->current_task != NULL) {
@@ -285,6 +294,7 @@ JL_DLLEXPORT void jl_atexit_hook(int exitcode)
     // force libuv to spin until everything has finished closing
     loop->stop_flag = 0;
     while (uv_run(loop, UV_RUN_DEFAULT)) { }
+    JL_UV_UNLOCK();
 
     // TODO: Destroy threads
 
@@ -294,8 +304,7 @@ JL_DLLEXPORT void jl_atexit_hook(int exitcode)
 #endif
 }
 
-void jl_get_builtin_hooks(void);
-void jl_get_builtins(void);
+static void post_boot_hooks(void);
 
 JL_DLLEXPORT void *jl_dl_handle;
 void *jl_RTLD_DEFAULT_handle;
@@ -384,6 +393,7 @@ static void *init_stdio_handle(const char *stdio, uv_os_fd_t fd, int readable)
         {
             int nullfd;
             nullfd = open("/dev/null", O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH /* 0666 */);
+            assert(nullfd != -1);
             dup2(nullfd, fd);
             close(nullfd);
         }
@@ -614,11 +624,6 @@ static void jl_resolve_sysimg_location(JL_IMAGE_SEARCH rel)
         jl_options.outputbc = abspath(jl_options.outputbc, 0);
     if (jl_options.machine_file)
         jl_options.machine_file = abspath(jl_options.machine_file, 0);
-    if (jl_options.project
-            && strcmp(jl_options.project, "@.") != 0
-            && strcmp(jl_options.project, "@") != 0
-            && strcmp(jl_options.project, "") != 0)
-        jl_options.project = abspath(jl_options.project, 0);
     if (jl_options.output_code_coverage)
         jl_options.output_code_coverage = absformat(jl_options.output_code_coverage);
 
@@ -642,10 +647,8 @@ static void jl_set_io_wait(int v)
 void _julia_init(JL_IMAGE_SEARCH rel)
 {
     jl_init_timing();
-#ifdef JULIA_ENABLE_THREADING
     // Make sure we finalize the tls callback before starting any threads.
     jl_get_ptls_states_getter();
-#endif
     jl_ptls_t ptls = jl_get_ptls_states();
     (void)ptls; assert(ptls); // make sure early that we have initialized ptls
     jl_safepoint_init();
@@ -654,16 +657,10 @@ void _julia_init(JL_IMAGE_SEARCH rel)
     ios_set_io_wait_func = jl_set_io_wait;
     jl_io_loop = uv_default_loop(); // this loop will internal events (spawning process etc.),
                                     // best to call this first, since it also initializes libuv
-    jl_init_signal_async();
+    jl_init_uv();
+    init_stdio();
     restore_signals();
 
-    jl_resolve_sysimg_location(rel);
-    // loads sysimg if available, and conditionally sets jl_options.cpu_target
-    if (jl_options.image_file)
-        jl_preload_sysimg_so(jl_options.image_file);
-    if (jl_options.cpu_target == NULL)
-        jl_options.cpu_target = "native";
-
     jl_page_size = jl_getpagesize();
     uint64_t total_mem = uv_get_total_memory();
     if (total_mem >= (size_t)-1) {
@@ -734,79 +731,53 @@ void _julia_init(JL_IMAGE_SEARCH rel)
     }
 #endif
 
-    jl_init_threading();
-
-    jl_gc_init();
-    jl_gc_enable(0);
-    jl_init_types();
-    jl_init_frontend();
-    jl_init_tasks();
-    jl_init_root_task(stack_lo, stack_hi);
-
-#ifdef ENABLE_TIMINGS
-    jl_root_task->timing_stack = jl_root_timing;
-#endif
-
-    init_stdio();
-    // libuv stdio cleanup depends on jl_init_tasks() because JL_TRY is used in jl_atexit_hook()
-
     if ((jl_options.outputo || jl_options.outputbc) &&
         (jl_options.code_coverage || jl_options.malloc_log)) {
         jl_error("cannot generate code-coverage or track allocation information while generating a .o or .bc output file");
     }
 
-    jl_init_codegen();
+    jl_gc_init();
 
-    jl_start_threads();
-
-    jl_an_empty_vec_any = (jl_value_t*)jl_alloc_vec_any(0);
-    jl_init_serializer();
+    jl_init_threading();
     jl_init_intrinsic_properties();
 
+    jl_gc_enable(0);
+
+    jl_resolve_sysimg_location(rel);
+    // loads sysimg if available, and conditionally sets jl_options.cpu_target
+    if (jl_options.image_file)
+        jl_preload_sysimg_so(jl_options.image_file);
+    if (jl_options.cpu_target == NULL)
+        jl_options.cpu_target = "native";
+
+    arraylist_new(&partial_inst, 0);
+    if (jl_options.image_file) {
+        jl_restore_system_image(jl_options.image_file);
+    }
+    else {
+        jl_init_types();
+        jl_init_codegen();
+    }
+
+    jl_init_tasks();
+    jl_init_root_task(stack_lo, stack_hi);
+#ifdef ENABLE_TIMINGS
+    jl_root_task->timing_stack = jl_root_timing;
+#endif
+    jl_init_frontend();
+
+    jl_an_empty_vec_any = (jl_value_t*)jl_alloc_vec_any(0); // used by ml_matches
+    jl_init_serializer();
+
     if (!jl_options.image_file) {
         jl_core_module = jl_new_module(jl_symbol("Core"));
         jl_type_typename->mt->module = jl_core_module;
         jl_top_module = jl_core_module;
         jl_init_intrinsic_functions();
         jl_init_primitives();
-        jl_get_builtins();
         jl_init_main_module();
-
         jl_load(jl_core_module, "boot.jl");
-        jl_get_builtin_hooks();
-        jl_init_box_caches();
-    }
-
-    if (jl_options.image_file) {
-        JL_TRY {
-            jl_restore_system_image(jl_options.image_file);
-        }
-        JL_CATCH {
-            jl_printf(JL_STDERR, "error during init:\n");
-            jl_static_show(JL_STDERR, jl_current_exception());
-            jl_printf(JL_STDERR, "\n");
-            jl_exit(1);
-        }
-    }
-
-    // set module field of primitive types
-    int i;
-    void **table = jl_core_module->bindings.table;
-    for(i=1; i < jl_core_module->bindings.size; i+=2) {
-        if (table[i] != HT_NOTFOUND) {
-            jl_binding_t *b = (jl_binding_t*)table[i];
-            jl_value_t *v = b->value;
-            if (v) {
-                if (jl_is_unionall(v))
-                    v = jl_unwrap_unionall(v);
-                if (jl_is_datatype(v)) {
-                    jl_datatype_t *tt = (jl_datatype_t*)v;
-                    tt->name->module = jl_core_module;
-                    if (tt->name->mt)
-                        tt->name->mt->module = jl_core_module;
-                }
-            }
-        }
+        post_boot_hooks();
     }
 
     // the Main module is the one which is always open, and set as the
@@ -814,7 +785,25 @@ void _julia_init(JL_IMAGE_SEARCH rel)
     // it does "using Base" if Base is available.
     if (jl_base_module != NULL) {
         jl_add_standard_imports(jl_main_module);
+        jl_value_t *maininclude = jl_get_global(jl_base_module, jl_symbol("MainInclude"));
+        if (maininclude && jl_is_module(maininclude)) {
+            jl_module_import(jl_main_module, (jl_module_t*)maininclude, jl_symbol("include"));
+            jl_module_import(jl_main_module, (jl_module_t*)maininclude, jl_symbol("eval"));
+        }
+        // Do initialization needed before starting child threads
+        jl_value_t *f = jl_get_global(jl_base_module, jl_symbol("__preinit_threads__"));
+        if (f) {
+            size_t last_age = ptls->world_age;
+            ptls->world_age = jl_get_world_counter();
+            jl_apply(&f, 1);
+            ptls->world_age = last_age;
+        }
     }
+    else {
+        // nthreads > 1 requires code in Base
+        jl_n_threads = 1;
+    }
+    jl_start_threads();
 
     // This needs to be after jl_start_threads
     if (jl_options.handle_signals == JL_OPTIONS_HANDLE_SIGNALS_ON)
@@ -844,30 +833,21 @@ static jl_value_t *core(const char *name)
 }
 
 // fetch references to things defined in boot.jl
-void jl_get_builtin_hooks(void)
+static void post_boot_hooks(void)
 {
-    int t;
-    for (t = 0; t < jl_n_threads; t++) {
-        jl_ptls_t ptls2 = jl_all_tls_states[t];
-        ptls2->root_task->tls = jl_nothing;
-        ptls2->root_task->donenotify = jl_nothing;
-        ptls2->root_task->exception = jl_nothing;
-        ptls2->root_task->result = jl_nothing;
-    }
-
     jl_char_type    = (jl_datatype_t*)core("Char");
     jl_int8_type    = (jl_datatype_t*)core("Int8");
     jl_int16_type   = (jl_datatype_t*)core("Int16");
     jl_uint16_type  = (jl_datatype_t*)core("UInt16");
-
     jl_float16_type = (jl_datatype_t*)core("Float16");
     jl_float32_type = (jl_datatype_t*)core("Float32");
     jl_float64_type = (jl_datatype_t*)core("Float64");
     jl_floatingpoint_type = (jl_datatype_t*)core("AbstractFloat");
-    jl_number_type = (jl_datatype_t*)core("Number");
-    jl_signed_type = (jl_datatype_t*)core("Signed");
+    jl_number_type  = (jl_datatype_t*)core("Number");
+    jl_signed_type  = (jl_datatype_t*)core("Signed");
     jl_datatype_t *jl_unsigned_type = (jl_datatype_t*)core("Unsigned");
     jl_datatype_t *jl_integer_type = (jl_datatype_t*)core("Integer");
+
     jl_bool_type->super = jl_integer_type;
     jl_uint8_type->super = jl_unsigned_type;
     jl_int32_type->super = jl_signed_type;
@@ -884,35 +864,39 @@ void jl_get_builtin_hooks(void)
     jl_boundserror_type    = (jl_datatype_t*)core("BoundsError");
     jl_memory_exception    = jl_new_struct_uninit((jl_datatype_t*)core("OutOfMemoryError"));
     jl_readonlymemory_exception = jl_new_struct_uninit((jl_datatype_t*)core("ReadOnlyMemoryError"));
-    jl_typeerror_type = (jl_datatype_t*)core("TypeError");
-
+    jl_typeerror_type      = (jl_datatype_t*)core("TypeError");
 #ifdef SEGV_EXCEPTION
     jl_segv_exception      = jl_new_struct_uninit((jl_datatype_t*)core("SegmentationFault"));
 #endif
+    jl_argumenterror_type  = (jl_datatype_t*)core("ArgumentError");
+    jl_methoderror_type    = (jl_datatype_t*)core("MethodError");
+    jl_loaderror_type      = (jl_datatype_t*)core("LoadError");
+    jl_initerror_type      = (jl_datatype_t*)core("InitError");
 
     jl_weakref_type = (jl_datatype_t*)core("WeakRef");
     jl_vecelement_typename = ((jl_datatype_t*)jl_unwrap_unionall(core("VecElement")))->name;
 
-    jl_argumenterror_type = (jl_datatype_t*)core("ArgumentError");
-    jl_methoderror_type = (jl_datatype_t*)core("MethodError");
-    jl_loaderror_type = (jl_datatype_t*)core("LoadError");
-    jl_initerror_type = (jl_datatype_t*)core("InitError");
-}
+    jl_init_box_caches();
 
-void jl_get_builtins(void)
-{
-    jl_builtin_throw = core("throw");           jl_builtin_is = core("===");
-    jl_builtin_typeof = core("typeof");         jl_builtin_sizeof = core("sizeof");
-    jl_builtin_issubtype = core("<:");          jl_builtin_isa = core("isa");
-    jl_builtin_typeassert = core("typeassert"); jl_builtin__apply = core("_apply");
-    jl_builtin_isdefined = core("isdefined");   jl_builtin_nfields = core("nfields");
-    jl_builtin_tuple = core("tuple");           jl_builtin_svec = core("svec");
-    jl_builtin_getfield = core("getfield");     jl_builtin_setfield = core("setfield!");
-    jl_builtin_fieldtype = core("fieldtype");   jl_builtin_arrayref = core("arrayref");
-    jl_builtin_arrayset = core("arrayset");     jl_builtin_arraysize = core("arraysize");
-    jl_builtin_apply_type = core("apply_type"); jl_builtin_applicable = core("applicable");
-    jl_builtin_invoke = core("invoke");         jl_builtin__expr = core("_expr");
-    jl_builtin_ifelse = core("ifelse");
+    // set module field of primitive types
+    int i;
+    void **table = jl_core_module->bindings.table;
+    for (i = 1; i < jl_core_module->bindings.size; i += 2) {
+        if (table[i] != HT_NOTFOUND) {
+            jl_binding_t *b = (jl_binding_t*)table[i];
+            jl_value_t *v = b->value;
+            if (v) {
+                if (jl_is_unionall(v))
+                    v = jl_unwrap_unionall(v);
+                if (jl_is_datatype(v)) {
+                    jl_datatype_t *tt = (jl_datatype_t*)v;
+                    tt->name->module = jl_core_module;
+                    if (tt->name->mt)
+                        tt->name->mt->module = jl_core_module;
+                }
+            }
+        }
+    }
 }
 
 #ifdef __cplusplus
