@@ -281,15 +281,15 @@ static Value *emit_unboxed_coercion(jl_codectx_t &ctx, Type *to, Value *unboxed)
     Type *ty = unboxed->getType();
     bool frompointer = ty->isPointerTy();
     bool topointer = to->isPointerTy();
-#if JL_LLVM_VERSION >= 40000
     const DataLayout &DL = jl_data_layout;
-#else
-    const DataLayout &DL = jl_ExecutionEngine->getDataLayout();
-#endif
     if (ty == T_int1 && to == T_int8) {
         // bools may be stored internally as int8
         unboxed = ctx.builder.CreateZExt(unboxed, T_int8);
     }
+    else if (ty == T_int8 && to == T_int1) {
+        // bools may be stored internally as int8
+        unboxed = ctx.builder.CreateTrunc(unboxed, T_int1);
+    }
     else if (ty == T_void || DL.getTypeSizeInBits(ty) != DL.getTypeSizeInBits(to)) {
         // this can happen in dead code
         //emit_unreachable(ctx);
@@ -381,11 +381,7 @@ static Value *emit_unbox(jl_codectx_t &ctx, Type *to, const jl_cgval_t &x, jl_va
             // appropriate coercion manually.
             AllocaInst *AI = cast<AllocaInst>(p);
             Type *AllocType = AI->getAllocatedType();
-#if JL_LLVM_VERSION >= 40000
             const DataLayout &DL = jl_data_layout;
-#else
-            const DataLayout &DL = jl_ExecutionEngine->getDataLayout();
-#endif
             if (!AI->isArrayAllocation() &&
                     (AllocType->isFloatingPointTy() || AllocType->isIntegerTy() || AllocType->isPointerTy()) &&
                     (to->isFloatingPointTy() || to->isIntegerTy() || to->isPointerTy()) &&
@@ -510,7 +506,7 @@ static jl_cgval_t generic_bitcast(jl_codectx_t &ctx, const jl_cgval_t *argv)
 
 static jl_cgval_t generic_cast(
         jl_codectx_t &ctx,
-        intrinsic f, Value *(*generic)(jl_codectx_t&, Type*, Value*),
+        intrinsic f, Instruction::CastOps Op,
         const jl_cgval_t *argv, bool toint, bool fromint)
 {
     const jl_cgval_t &targ = argv[0];
@@ -531,64 +527,23 @@ static jl_cgval_t generic_cast(
     if (!to || !vt)
         return emit_runtime_call(ctx, f, argv, 2);
     Value *from = emit_unbox(ctx, vt, v, v.typ);
-    Value *ans = generic(ctx, to, from);
-    return mark_julia_type(ctx, ans, false, jlto);
-}
-
-static Value *generic_trunc(jl_codectx_t &ctx, Type *to, Value *x)
-{
-    return ctx.builder.CreateTrunc(x, to);
-}
-
-static Value *generic_sext(jl_codectx_t &ctx, Type *to, Value *x)
-{
-    return ctx.builder.CreateSExt(x, to);
-}
-
-static Value *generic_zext(jl_codectx_t &ctx, Type *to, Value *x)
-{
-    return ctx.builder.CreateZExt(x, to);
-}
-
-static Value *generic_uitofp(jl_codectx_t &ctx, Type *to, Value *x)
-{
-    return ctx.builder.CreateUIToFP(x, to);
-}
-
-static Value *generic_sitofp(jl_codectx_t &ctx, Type *to, Value *x)
-{
-    return ctx.builder.CreateSIToFP(x, to);
-}
-
-static Value *generic_fptoui(jl_codectx_t &ctx, Type *to, Value *x)
-{
-    return ctx.builder.CreateFPToUI(x, to);
-}
-
-static Value *generic_fptosi(jl_codectx_t &ctx, Type *to, Value *x)
-{
-    return ctx.builder.CreateFPToSI(x, to);
-}
-
-static Value *generic_fptrunc(jl_codectx_t &ctx, Type *to, Value *x)
-{
-    return ctx.builder.CreateFPTrunc(x, to);
-}
-
-static Value *generic_fpext(jl_codectx_t &ctx, Type *to, Value *x)
-{
+    if (!CastInst::castIsValid(Op, from, to))
+        return emit_runtime_call(ctx, f, argv, 2);
+    if (Op == Instruction::FPExt) {
 #ifdef JL_NEED_FLOATTEMP_VAR
-    // Target platform might carry extra precision.
-    // Force rounding to single precision first. The reason is that it's
-    // fine to keep working in extended precision as long as it's
-    // understood that everything is implicitly rounded to 23 bits,
-    // but if we start looking at more bits we need to actually do the
-    // rounding first instead of carrying around incorrect low bits.
-    Value *jlfloattemp_var = emit_static_alloca(ctx, x->getType());
-    ctx.builder.CreateStore(x, jlfloattemp_var);
-    x  = ctx.builder.CreateLoad(jlfloattemp_var, true);
+        // Target platform might carry extra precision.
+        // Force rounding to single precision first. The reason is that it's
+        // fine to keep working in extended precision as long as it's
+        // understood that everything is implicitly rounded to 23 bits,
+        // but if we start looking at more bits we need to actually do the
+        // rounding first instead of carrying around incorrect low bits.
+        Value *jlfloattemp_var = emit_static_alloca(ctx, from->getType());
+        ctx.builder.CreateStore(from, jlfloattemp_var);
+        from  = ctx.builder.CreateLoad(jlfloattemp_var, /*force this to load from the stack*/true);
 #endif
-    return ctx.builder.CreateFPExt(x, to);
+    }
+    Value *ans = ctx.builder.CreateCast(Op, from, to);
+    return mark_julia_type(ctx, ans, false, jlto);
 }
 
 static jl_cgval_t emit_runtime_pointerref(jl_codectx_t &ctx, jl_cgval_t *argv)
@@ -649,8 +604,13 @@ static jl_cgval_t emit_pointerref(jl_codectx_t &ctx, jl_cgval_t *argv)
         bool isboxed;
         Type *ptrty = julia_type_to_llvm(ety, &isboxed);
         assert(!isboxed);
-        Value *thePtr = emit_unbox(ctx, ptrty->getPointerTo(), e, e.typ);
-        return typed_load(ctx, thePtr, im1, ety, tbaa_data, true, align_nb);
+        if (!type_is_ghost(ptrty)) {
+            Value *thePtr = emit_unbox(ctx, ptrty->getPointerTo(), e, e.typ);
+            return typed_load(ctx, thePtr, im1, ety, tbaa_data, nullptr, true, align_nb);
+        }
+        else {
+            return ghostValue(ety);
+        }
     }
 }
 
@@ -683,7 +643,7 @@ static jl_cgval_t emit_pointerset(jl_codectx_t &ctx, jl_cgval_t *argv)
         return emit_runtime_pointerset(ctx, argv);
     if (!jl_is_datatype(ety))
         ety = (jl_value_t*)jl_any_type;
-    emit_typecheck(ctx, x, ety, "pointerset: type mismatch in assign");
+    emit_typecheck(ctx, x, ety, "pointerset");
 
     Value *idx = emit_unbox(ctx, T_size, i, (jl_value_t*)jl_long_type);
     Value *im1 = ctx.builder.CreateSub(idx, ConstantInt::get(T_size, 1));
@@ -712,10 +672,12 @@ static jl_cgval_t emit_pointerset(jl_codectx_t &ctx, jl_cgval_t *argv)
         bool isboxed;
         Type *ptrty = julia_type_to_llvm(ety, &isboxed);
         assert(!isboxed);
-        thePtr = emit_unbox(ctx, ptrty->getPointerTo(), e, e.typ);
-        typed_store(ctx, thePtr, im1, x, ety, tbaa_data, NULL, align_nb);
+        if (!type_is_ghost(ptrty)) {
+            thePtr = emit_unbox(ctx, ptrty->getPointerTo(), e, e.typ);
+            typed_store(ctx, thePtr, im1, x, ety, tbaa_data, nullptr, nullptr, align_nb);
+        }
     }
-    return mark_julia_type(ctx, thePtr, false, aty);
+    return e;
 }
 
 static Value *emit_checked_srem_int(jl_codectx_t &ctx, Value *x, Value *den)
@@ -755,18 +717,10 @@ struct math_builder {
         if (jl_options.fast_math != JL_OPTIONS_FAST_MATH_OFF &&
             (always_fast ||
              jl_options.fast_math == JL_OPTIONS_FAST_MATH_ON)) {
-#if JL_LLVM_VERSION >= 60000
             fmf.setFast();
-#else
-            fmf.setUnsafeAlgebra();
-#endif
         }
-#if JL_LLVM_VERSION >= 50000
         if (contract)
             fmf.setAllowContract(true);
-#else
-        assert(!contract);
-#endif
         ctxbuilder.setFastMathFlags(fmf);
     }
     IRBuilder<>& operator()() const { return ctxbuilder; }
@@ -833,18 +787,22 @@ static jl_cgval_t emit_ifelse(jl_codectx_t &ctx, jl_cgval_t c, jl_cgval_t x, jl_
             Value *y_vboxed = y.Vboxed;
             Value *x_ptr = (x.isghost ? NULL : data_pointer(ctx, x));
             Value *y_ptr = (y.isghost ? NULL : data_pointer(ctx, y));
+            MDNode *ifelse_tbaa;
             if (!x.isghost && x.constant)
                 x_vboxed = boxed(ctx, x);
             if (!y.isghost && y.constant)
                 y_vboxed = boxed(ctx, y);
-            if (!x_ptr && !y_ptr) {
+            if (!x_ptr && !y_ptr) { // both ghost
                 ifelse_result = NULL;
+                ifelse_tbaa = tbaa_stack;
             }
             else if (!x_ptr) {
                 ifelse_result = y_ptr;
+                ifelse_tbaa = y.tbaa;
             }
             else if (!y_ptr) {
                 ifelse_result = x_ptr;
+                ifelse_tbaa = x.tbaa;
             }
             else {
                 x_ptr = decay_derived(x_ptr);
@@ -852,6 +810,13 @@ static jl_cgval_t emit_ifelse(jl_codectx_t &ctx, jl_cgval_t c, jl_cgval_t x, jl_
                 if (x_ptr->getType() != y_ptr->getType())
                     y_ptr = ctx.builder.CreateBitCast(y_ptr, x_ptr->getType());
                 ifelse_result = ctx.builder.CreateSelect(isfalse, y_ptr, x_ptr);
+                ifelse_tbaa = MDNode::getMostGenericTBAA(x.tbaa, y.tbaa);
+                if (ifelse_tbaa == NULL) {
+                    // LLVM won't return a TBAA result for the root, but mark_julia_struct requires it: make it now
+                    auto *OffsetNode = ConstantAsMetadata::get(ConstantInt::get(T_int64, 0));
+                    Metadata *Ops[] = {tbaa_root, tbaa_root, OffsetNode};
+                    ifelse_tbaa = MDNode::get(jl_LLVMContext, Ops);
+                }
             }
             Value *tindex;
             if (!x_tindex && x.constant) {
@@ -889,7 +854,7 @@ static jl_cgval_t emit_ifelse(jl_codectx_t &ctx, jl_cgval_t c, jl_cgval_t x, jl_
                 ctx.builder.Insert(ret);
                 tindex = ret;
             }
-            jl_cgval_t ret = mark_julia_slot(ifelse_result, rt_hint, tindex, tbaa_data);
+            jl_cgval_t ret = mark_julia_slot(ifelse_result, rt_hint, tindex, ifelse_tbaa);
             if (x_vboxed || y_vboxed) {
                 if (!x_vboxed)
                     x_vboxed = ConstantPointerNull::get(cast<PointerType>(y_vboxed->getType()));
@@ -940,23 +905,23 @@ static jl_cgval_t emit_intrinsic(jl_codectx_t &ctx, intrinsic f, jl_value_t **ar
     case bitcast:
         return generic_bitcast(ctx, argv);
     case trunc_int:
-        return generic_cast(ctx, f, generic_trunc, argv, true, true);
+        return generic_cast(ctx, f, Instruction::Trunc, argv, true, true);
     case sext_int:
-        return generic_cast(ctx, f, generic_sext, argv, true, true);
+        return generic_cast(ctx, f, Instruction::SExt, argv, true, true);
     case zext_int:
-        return generic_cast(ctx, f, generic_zext, argv, true, true);
+        return generic_cast(ctx, f, Instruction::ZExt, argv, true, true);
     case uitofp:
-        return generic_cast(ctx, f, generic_uitofp, argv, false, true);
+        return generic_cast(ctx, f, Instruction::UIToFP, argv, false, true);
     case sitofp:
-        return generic_cast(ctx, f, generic_sitofp, argv, false, true);
+        return generic_cast(ctx, f, Instruction::SIToFP, argv, false, true);
     case fptoui:
-        return generic_cast(ctx, f, generic_fptoui, argv, true, false);
+        return generic_cast(ctx, f, Instruction::FPToUI, argv, true, false);
     case fptosi:
-        return generic_cast(ctx, f, generic_fptosi, argv, true, false);
+        return generic_cast(ctx, f, Instruction::FPToSI, argv, true, false);
     case fptrunc:
-        return generic_cast(ctx, f, generic_fptrunc, argv, false, false);
+        return generic_cast(ctx, f, Instruction::FPTrunc, argv, false, false);
     case fpext:
-        return generic_cast(ctx, f, generic_fpext, argv, false, false);
+        return generic_cast(ctx, f, Instruction::FPExt, argv, false, false);
 
     case not_int: {
         const jl_cgval_t &x = argv[0];
@@ -986,6 +951,15 @@ static jl_cgval_t emit_intrinsic(jl_codectx_t &ctx, intrinsic f, jl_value_t **ar
             xtyp = INTT(xtyp);
         if (!xtyp)
             return emit_runtime_call(ctx, f, argv, nargs);
+        ////Bool are required to be in the range [0,1]
+        ////so while they are represented as i8,
+        ////the operations need to be done in mod 1
+        ////we can either do that now, or truncate them
+        ////later into mod 1.
+        ////LLVM seems to emit better code if we do the latter,
+        ////(more likely to fold away the cast) so that's what we'll do.
+        //if (xtyp == (jl_value_t*)jl_bool_type)
+        //    r = T_int1;
 
         Type **argt = (Type**)alloca(sizeof(Type*) * nargs);
         argt[0] = xtyp;
@@ -1010,11 +984,12 @@ static jl_cgval_t emit_intrinsic(jl_codectx_t &ctx, intrinsic f, jl_value_t **ar
         }
 
         // call the intrinsic
-        jl_value_t *newtyp = NULL;
+        jl_value_t *newtyp = xinfo.typ;
         Value *r = emit_untyped_intrinsic(ctx, f, argvalues, nargs, (jl_datatype_t**)&newtyp, xinfo.typ);
-        if (r->getType() == T_int1)
-            r = ctx.builder.CreateZExt(r, T_int8);
-        return mark_julia_type(ctx, r, false, newtyp ? newtyp : xinfo.typ);
+        // Turn Bool operations into mod 1 now, if needed
+        if (newtyp == (jl_value_t*)jl_bool_type && r->getType() != T_int1)
+            r = ctx.builder.CreateTrunc(r, T_int1);
+        return mark_julia_type(ctx, r, false, newtyp);
     }
     }
     assert(0 && "unreachable");
@@ -1079,18 +1054,11 @@ static Value *emit_untyped_intrinsic(jl_codectx_t &ctx, intrinsic f, Value **arg
         return ctx.builder.CreateCall(fmaintr, {x, y, z});
     }
     case muladd_float: {
-#if JL_LLVM_VERSION >= 50000
         // LLVM 5.0 can create FMA in the backend for contractable fmul and fadd
         // Emitting fmul and fadd here since they are easier for other LLVM passes to
         // optimize.
         auto mathb = math_builder(ctx, false, true);
         return mathb().CreateFAdd(mathb().CreateFMul(x, y), z);
-#else
-        assert(y->getType() == x->getType());
-        assert(z->getType() == y->getType());
-        Value *muladdintr = Intrinsic::getDeclaration(jl_Module, Intrinsic::fmuladd, makeArrayRef(t));
-        return ctx.builder.CreateCall(muladdintr, {x, y, z});
-#endif
     }
 
     case checked_sadd_int:
