@@ -19,7 +19,6 @@
 
 static void attach_exception_port(thread_port_t thread, int segv_only);
 
-#ifdef JULIA_ENABLE_THREADING
 // low 16 bits are the thread id, the next 8 bits are the original gc_state
 static arraylist_t suspended_threads;
 void jl_mach_gc_end(void)
@@ -31,7 +30,7 @@ void jl_mach_gc_end(void)
         int8_t gc_state = (int8_t)(item >> 8);
         jl_ptls_t ptls2 = jl_all_tls_states[tid];
         jl_atomic_store_release(&ptls2->gc_state, gc_state);
-        thread_resume(pthread_mach_thread_np(ptls2->system_id));
+        thread_resume(pthread_mach_thread_np((pthread_t)ptls2->system_id));
     }
     suspended_threads.len = 0;
 }
@@ -57,7 +56,6 @@ static int jl_mach_gc_wait(jl_ptls_t ptls2,
     jl_mutex_unlock_nogc(&safepoint_lock);
     return 1;
 }
-#endif
 
 static mach_port_t segv_port = 0;
 
@@ -80,9 +78,7 @@ void *mach_segv_listener(void *arg)
 
 static void allocate_segv_handler()
 {
-#ifdef JULIA_ENABLE_THREADING
     arraylist_new(&suspended_threads, jl_n_threads);
-#endif
     pthread_t thread;
     pthread_attr_t attr;
     kern_return_t ret;
@@ -101,7 +97,7 @@ static void allocate_segv_handler()
     }
     pthread_attr_destroy(&attr);
     for (int16_t tid = 0;tid < jl_n_threads;tid++) {
-        attach_exception_port(pthread_mach_thread_np(jl_all_tls_states[tid]->system_id), 0);
+        attach_exception_port(pthread_mach_thread_np((pthread_t)jl_all_tls_states[tid]->system_id), 0);
     }
 }
 
@@ -147,9 +143,9 @@ static void jl_throw_in_thread(int tid, mach_port_t thread, jl_value_t *exceptio
         assert(exception);
         ptls2->bt_size = rec_backtrace_ctx(ptls2->bt_data, JL_MAX_BT_SIZE,
                                            (bt_context_t*)&state);
-        ptls2->exception_in_transit = exception;
+        ptls2->sig_exception = exception;
     }
-    jl_call_in_state(ptls2, &state, &jl_rethrow);
+    jl_call_in_state(ptls2, &state, &jl_sig_throw);
     ret = thread_set_state(thread, x86_THREAD_STATE64,
                            (thread_state_t)&state, count);
     HANDLE_MACH_ERROR("thread_set_state",ret);
@@ -174,11 +170,10 @@ kern_return_t catch_exception_raise(mach_port_t            exception_port,
     }
 #endif
     int16_t tid;
-#ifdef JULIA_ENABLE_THREADING
     jl_ptls_t ptls2 = NULL;
     for (tid = 0;tid < jl_n_threads;tid++) {
         jl_ptls_t _ptls2 = jl_all_tls_states[tid];
-        if (pthread_mach_thread_np(_ptls2->system_id) == thread) {
+        if (pthread_mach_thread_np((pthread_t)_ptls2->system_id) == thread) {
             ptls2 = _ptls2;
             break;
         }
@@ -190,10 +185,6 @@ kern_return_t catch_exception_raise(mach_port_t            exception_port,
         jl_safe_printf("ERROR: Exception handler triggered on unmanaged thread.\n");
         return KERN_INVALID_ARGUMENT;
     }
-#else
-    jl_ptls_t ptls2 = &jl_tls_states;
-    tid = 0;
-#endif
     if (exception == EXC_ARITHMETIC) {
         jl_throw_in_thread(tid, thread, jl_diverror_exception);
         return KERN_SUCCESS;
@@ -203,12 +194,10 @@ kern_return_t catch_exception_raise(mach_port_t            exception_port,
     HANDLE_MACH_ERROR("thread_get_state", ret);
     uint64_t fault_addr = exc_state.__faultvaddr;
     if (jl_addr_is_safepoint(fault_addr)) {
-#ifdef JULIA_ENABLE_THREADING
         if (jl_mach_gc_wait(ptls2, thread, tid))
             return KERN_SUCCESS;
         if (ptls2->tid != 0)
             return KERN_SUCCESS;
-#endif
         if (ptls2->defer_signal) {
             jl_safepoint_defer_sigint();
         }
@@ -269,7 +258,7 @@ static void attach_exception_port(thread_port_t thread, int segv_only)
 static void jl_thread_suspend_and_get_state(int tid, unw_context_t **ctx)
 {
     jl_ptls_t ptls2 = jl_all_tls_states[tid];
-    mach_port_t tid_port = pthread_mach_thread_np(ptls2->system_id);
+    mach_port_t tid_port = pthread_mach_thread_np((pthread_t)ptls2->system_id);
 
     kern_return_t ret = thread_suspend(tid_port);
     HANDLE_MACH_ERROR("thread_suspend", ret);
@@ -289,7 +278,7 @@ static void jl_thread_suspend_and_get_state(int tid, unw_context_t **ctx)
 static void jl_thread_resume(int tid, int sig)
 {
     jl_ptls_t ptls2 = jl_all_tls_states[tid];
-    mach_port_t thread = pthread_mach_thread_np(ptls2->system_id);
+    mach_port_t thread = pthread_mach_thread_np((pthread_t)ptls2->system_id);
     kern_return_t ret = thread_resume(thread);
     HANDLE_MACH_ERROR("thread_resume", ret);
 }
@@ -299,7 +288,7 @@ static void jl_thread_resume(int tid, int sig)
 static void jl_try_deliver_sigint(void)
 {
     jl_ptls_t ptls2 = jl_all_tls_states[0];
-    mach_port_t thread = pthread_mach_thread_np(ptls2->system_id);
+    mach_port_t thread = pthread_mach_thread_np((pthread_t)ptls2->system_id);
 
     kern_return_t ret = thread_suspend(thread);
     HANDLE_MACH_ERROR("thread_suspend", ret);
@@ -328,7 +317,7 @@ static void jl_try_deliver_sigint(void)
 static void jl_exit_thread0(int exitstate)
 {
     jl_ptls_t ptls2 = jl_all_tls_states[0];
-    mach_port_t thread = pthread_mach_thread_np(ptls2->system_id);
+    mach_port_t thread = pthread_mach_thread_np((pthread_t)ptls2->system_id);
     kern_return_t ret = thread_suspend(thread);
     HANDLE_MACH_ERROR("thread_suspend", ret);
 
