@@ -34,7 +34,7 @@ end
 ####
 
 """
-    init(; n::Integer, delay::Float64)
+    init(; n::Integer, delay::Real))
 
 Configure the `delay` between backtraces (measured in seconds), and the number `n` of
 instruction pointers that may be stored. Each instruction pointer corresponds to a single
@@ -42,7 +42,7 @@ line of code; backtraces generally consist of a long list of instruction pointer
 settings can be obtained by calling this function with no arguments, and each can be set
 independently using keywords or in the order `(n, delay)`.
 """
-function init(; n::Union{Nothing,Integer} = nothing, delay::Union{Nothing,Float64} = nothing)
+function init(; n::Union{Nothing,Integer} = nothing, delay::Union{Nothing,Real} = nothing)
     n_cur = ccall(:jl_profile_maxlen_data, Csize_t, ())
     delay_cur = ccall(:jl_profile_delay_nsec, UInt64, ())/10^9
     if n === nothing && delay === nothing
@@ -53,7 +53,7 @@ function init(; n::Union{Nothing,Integer} = nothing, delay::Union{Nothing,Float6
     init(nnew, delaynew)
 end
 
-function init(n::Integer, delay::Float64)
+function init(n::Integer, delay::Real)
     status = ccall(:jl_profile_init, Cint, (Csize_t, UInt64), n, round(UInt64,10^9*delay))
     if status == -1
         error("could not allocate space for ", n, " instruction pointers")
@@ -143,9 +143,9 @@ end
 function print(io::IO, data::Vector{<:Unsigned}, lidict::Union{LineInfoDict, LineInfoFlatDict}, fmt::ProfileFormat, format::Symbol)
     cols::Int = Base.displaysize(io)[2]
     data = convert(Vector{UInt64}, data)
-    if format == :tree
+    if format === :tree
         tree(io, data, lidict, cols, fmt)
-    elseif format == :flat
+    elseif format === :flat
         flat(io, data, lidict, cols, fmt)
     else
         throw(ArgumentError("output format $(repr(format)) not recognized"))
@@ -360,11 +360,11 @@ end
 
 function flat(io::IO, data::Vector{UInt64}, lidict::Union{LineInfoDict, LineInfoFlatDict}, cols::Int, fmt::ProfileFormat)
     iplist, n = count_flat(data)
+    lilist, n = parse_flat(iplist, n, lidict, fmt.C)
     if isempty(n)
         warning_empty()
         return
     end
-    lilist, n = parse_flat(iplist, n, lidict, fmt.C)
     print_flat(io, lilist, n, cols, fmt)
     nothing
 end
@@ -390,7 +390,7 @@ function print_flat(io::IO, lilist::Vector{StackFrame}, n::Vector{Int},
         n = n[keep]
         lilist = lilist[keep]
     end
-    if fmt.sortedby == :count
+    if fmt.sortedby === :count
         p = sortperm(n)
         n = n[p]
         lilist = lilist[p]
@@ -514,6 +514,7 @@ function tree!(root::StackFrameTree{T}, all::Vector{UInt64}, lidict::Union{LineI
                 # jump forward to the end of the inlining chain
                 # avoiding an extra (slow) lookup of `ip` in `lidict`
                 # and an extra chain of them in `down`
+                # note that we may even have this === parent (if we're ignoring this frame ip)
                 this = builder_value[fastkey]
                 let this = this
                     while this !== parent
@@ -532,8 +533,7 @@ function tree!(root::StackFrameTree{T}, all::Vector{UInt64}, lidict::Union{LineI
                 frame = (frames isa Vector ? frames[i] : frames)
                 !C && frame.from_c && continue
                 key = (T === UInt64 ? ip : frame)
-                down = parent.down
-                this = get!(down, key) do
+                this = get!(parent.down, key) do
                     return StackFrameTree{T}()
                 end
                 this.frame = frame
@@ -547,37 +547,46 @@ function tree!(root::StackFrameTree{T}, all::Vector{UInt64}, lidict::Union{LineI
         end
     end
     function cleanup!(node::StackFrameTree)
-        empty!(node.builder_key)
-        empty!(node.builder_value)
-        foreach(cleanup!, values(node.down))
+        stack = StackFrameTree[node]
+        while !isempty(stack)
+            node = pop!(stack)
+            empty!(node.builder_key)
+            empty!(node.builder_value)
+            append!(stack, values(node.down))
+        end
         nothing
     end
     cleanup!(root)
     return root
 end
 
-# Print a "branch" starting at a particular level. This gets called recursively.
-function tree(io::IO, bt::StackFrameTree, level::Int, cols::Int, fmt::ProfileFormat, noisefloor::Int)
-    level > fmt.maxdepth && return
-    isempty(bt.down) && return
-    # Order the line information
-    nexts = collect(values(bt.down))
-    lilist = collect(frame.frame for frame in nexts)
-    counts = collect(frame.count for frame in nexts)
-    # Generate the string for each line
-    strs = tree_format(lilist, counts, level, cols)
-    # Recurse to the next level
-    for i in liperm(lilist)
-        down = nexts[i]
-        count = down.count
-        count < fmt.mincount && continue
-        count < noisefloor && continue
-        str = strs[i]
-        println(io, isempty(str) ? "$count unknown stackframe" : str)
-        noisefloor_down = fmt.noisefloor > 0 ? floor(Int, fmt.noisefloor * sqrt(count)) : 0
-        tree(io, down, level + 1, cols, fmt, noisefloor_down)
+# Print the stack frame tree starting at a particular root. Uses a worklist to
+# avoid stack overflows.
+function tree(io::IO, bt::StackFrameTree, cols::Int, fmt::ProfileFormat)
+    worklist = [(bt, 0, 0, "")]
+    while !isempty(worklist)
+        (bt, level, noisefloor, str) = popfirst!(worklist)
+        isempty(str) || println(io, str)
+        level > fmt.maxdepth && continue
+        isempty(bt.down) && continue
+        # Order the line information
+        nexts = collect(values(bt.down))
+        lilist = collect(frame.frame for frame in nexts)
+        counts = collect(frame.count for frame in nexts)
+        # Generate the string for each line
+        strs = tree_format(lilist, counts, level, cols)
+        # Recurse to the next level
+        for i in reverse(liperm(lilist))
+            down = nexts[i]
+            count = down.count
+            count < fmt.mincount && continue
+            count < noisefloor && continue
+            str = strs[i]
+            isempty(str) && (str = "$count unknown stackframe")
+            noisefloor_down = fmt.noisefloor > 0 ? floor(Int, fmt.noisefloor * sqrt(count)) : 0
+            pushfirst!(worklist, (down, level + 1, noisefloor_down, str))
+        end
     end
-    nothing
 end
 
 function tree(io::IO, data::Vector{UInt64}, lidict::Union{LineInfoFlatDict, LineInfoDict}, cols::Int, fmt::ProfileFormat)
@@ -590,8 +599,7 @@ function tree(io::IO, data::Vector{UInt64}, lidict::Union{LineInfoFlatDict, Line
         warning_empty()
         return
     end
-    level = 0
-    tree(io, root, level, cols, fmt, 0)
+    tree(io, root, cols, fmt)
     nothing
 end
 
