@@ -338,7 +338,7 @@ end
 # will show it consist of Expr, QuoteNode's and Symbol's which all needs to
 # be handled differently to iterate down to get the value of whitespace_chars.
 function get_value(sym::Expr, fn)
-    sym.head != :. && return (nothing, false)
+    sym.head !== :. && return (nothing, false)
     for ex in sym.args
         fn, found = get_value(ex, fn)
         !found && return (nothing, false)
@@ -375,7 +375,7 @@ function get_type_call(expr::Expr)
         found ? push!(args, typ) : push!(args, Any)
     end
     # use _methods_by_ftype as the function is supplied as a type
-    world = ccall(:jl_get_world_counter, UInt, ())
+    world = Base.get_world_counter()
     mt = Base._methods_by_ftype(Tuple{ft, args...}, -1, world)
     length(mt) == 1 || return (Any, false)
     m = first(mt)
@@ -530,14 +530,11 @@ function dict_identifier_key(str,tag)
         sym = Symbol(name)
         isdefined(obj, sym) || return (nothing, nothing, nothing)
         obj = getfield(obj, sym)
-        # Avoid `isdefined(::Array, ::Symbol)`
-        isa(obj, Array) && return (nothing, nothing, nothing)
     end
-    begin_of_key = first(something(findnext(r"\S", str, nextind(str, end_of_identifier) + 1), 1)) # 1 for [
-    begin_of_key==0 && return (true, nothing, nothing)
-    partial_key = str[begin_of_key:end]
-    (isa(obj, AbstractDict) && length(obj) < 1e6) || return (true, nothing, nothing)
-    return (obj, partial_key, begin_of_key)
+    (isa(obj, AbstractDict) && length(obj) < 1_000_000) || return (nothing, nothing, nothing)
+    begin_of_key = something(findnext(!isspace, str, nextind(str, end_of_identifier) + 1), # +1 for [
+                             lastindex(str)+1)
+    return (obj, str[begin_of_key:end], begin_of_key)
 end
 
 # This needs to be a separate non-inlined function, see #19441
@@ -555,20 +552,18 @@ function project_deps_get_completion_candidates(pkgstarts::String, project_file:
     open(project_file) do io
         state = :top
         for line in eachline(io)
-            if state == :top
-                if occursin(Base.re_section, line)
-                    state = occursin(Base.re_section_deps, line) ? :deps : :other
-                elseif (m = match(Base.re_name_to_string, line)) != nothing
+            if occursin(Base.re_section, line)
+                state = occursin(Base.re_section_deps, line) ? :deps : :other
+            elseif state === :top
+                if (m = match(Base.re_name_to_string, line)) !== nothing
                     root_name = String(m.captures[1])
                     startswith(root_name, pkgstarts) && push!(loading_candidates, root_name)
                 end
-            elseif state == :deps
-                if (m = match(Base.re_key_to_string, line)) != nothing
+            elseif state === :deps
+                if (m = match(Base.re_key_to_string, line)) !== nothing
                     dep_name = m.captures[1]
                     startswith(dep_name, pkgstarts) && push!(loading_candidates, dep_name)
                 end
-            elseif occursin(Base.re_section, line)
-                state = occursin(Base.re_section_deps, line) ? :deps : :other
             end
         end
     end
@@ -583,13 +578,9 @@ function completions(string, pos, context_module=Main)::Completions
     # if completing a key in a Dict
     identifier, partial_key, loc = dict_identifier_key(partial,inc_tag)
     if identifier !== nothing
-        if partial_key !== nothing
-            matches = find_dict_matches(identifier, partial_key)
-            length(matches)==1 && (length(string) <= pos || string[pos+1] != ']') && (matches[1]*="]")
-            length(matches)>0 && return [DictCompletion(identifier, match) for match in sort!(matches)], loc:pos, true
-        else
-            return Completion[], 0:-1, false
-        end
+        matches = find_dict_matches(identifier, partial_key)
+        length(matches)==1 && (lastindex(string) <= pos || string[nextind(string,pos)] != ']') && (matches[1]*=']')
+        length(matches)>0 && return [DictCompletion(identifier, match) for match in sort!(matches)], loc:pos, true
     end
 
     # otherwise...
@@ -603,11 +594,11 @@ function completions(string, pos, context_module=Main)::Completions
 
         paths, r, success = complete_path(replace(string[r], r"\\ " => " "), pos)
 
-        if inc_tag == :string &&
+        if inc_tag === :string &&
            length(paths) == 1 &&  # Only close if there's a single choice,
            !isdir(expanduser(replace(string[startpos:prevind(string, first(r))] * paths[1].path,
                                      r"\\ " => " "))) &&  # except if it's a directory
-           (length(string) <= pos ||
+           (lastindex(string) <= pos ||
             string[nextind(string,pos)] != '"')  # or there's already a " at the cursor.
             paths[1] = PathCompletion(paths[1].path * "\"")
         end
@@ -621,7 +612,7 @@ function completions(string, pos, context_module=Main)::Completions
 
     # Make sure that only bslash_completions is working on strings
     inc_tag==:string && return String[], 0:-1, false
-    if inc_tag == :other && should_method_complete(partial)
+    if inc_tag === :other && should_method_complete(partial)
         frange, method_name_end = find_start_brace(partial)
         ex = Meta.parse(partial[frange] * ")", raise=false, depwarn=false)
 
@@ -632,7 +623,7 @@ function completions(string, pos, context_module=Main)::Completions
                 return complete_methods(ex, context_module), first(frange):(method_name_end - 1), false
             end
         end
-    elseif inc_tag == :comment
+    elseif inc_tag === :comment
         return Completion[], 0:-1, false
     end
 
